from tkinter import *
from typing import Optional

from simba.mixins.pop_up_mixin import PopUpMixin
#from simba.ui.create_project_ui.ProjectCreatorPopUp
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        FileSelect, SimbaButton, SimBALabel)
from simba.ui.user_defined_pose_creator import PoseConfigCreator
from simba.utils.checks import (check_file_exist_and_readable, check_int,
                                check_str, check_valid_img_path)
from simba.utils.enums import Formats, Options
from simba.utils.errors import DuplicationError, InvalidInputError
from simba.utils.printing import stdout_success


class CreateUserDefinedPoseConfigurationPopUp(PopUpMixin):
    def __init__(self,
                 master: Optional[Toplevel] = None,
                 project_config_class = None): # simba.ui.create_project_ui.ProjectCreatorPopUp

        PopUpMixin.__init__(self, title="USER-DEFINED POSE CONFIGURATION", size=(700, 700), main_scrollbar=True, icon='pose')
        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header='SETTINGS', icon_name='settings', padx=5, pady=5, relief='solid')
        self.config_name_entry_box = Entry_Box(parent=self.settings_frm, fileDescription="POSE CONFIG NAME: ", labelwidth=35, entry_box_width=40)
        self.animal_cnt_entry_box = Entry_Box(parent=self.settings_frm, fileDescription="# ANIMALS: ", labelwidth=35, entry_box_width=40, validation='numeric')
        self.no_body_parts_entry_box = Entry_Box(parent=self.settings_frm, fileDescription="# OF BODY-PARTS (PER ANIMAL): ", labelwidth=35, entry_box_width=40, validation='numeric')
        self.img_path_file_select = FileSelect(self.settings_frm, fileDescription="IMAGE PATH: ", lblwidth=35, file_types=[("IMAGE FILE", Options.ALL_IMAGE_FORMAT_OPTIONS.value)], entry_width=40)
        self.confirm_btn = SimbaButton(parent=self.settings_frm, txt=f"CONFIRM", img='tick', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.create_bodypart_table)
        self.save_btn = SimbaButton(parent=self.settings_frm, txt="SAVE USER-DEFINED POSE-CONFIG", img='rocket', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.save_pose_config)
        self.save_btn.config(state="disabled")
        self.master, self.project_config_class = master, project_config_class

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.config_name_entry_box.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_entry_box.grid(row=1, column=0, sticky=NW)
        self.no_body_parts_entry_box.grid(row=2, column=0, sticky=NW)
        self.img_path_file_select.grid(row=3, column=0, sticky=NW, pady=2)
        self.confirm_btn.grid(row=4, pady=5, column=0, sticky=NW)
        self.save_btn.grid(row=6, pady=5, column=0, sticky=NW)
        # self.root.lift()
        self.main_frm.mainloop()

    def create_bodypart_table(self):
        if hasattr(self, "bp_table_frm"):
            self.bp_table_frm.destroy()

        self.bp_table_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header='BODY-PARTS', icon_name='pose', padx=5, pady=5, relief='solid')

        check_int(name="ANIMAL NUMBER", value=self.animal_cnt_entry_box.entry_get)
        check_int(name="BODY-PART NUMBER", value=self.no_body_parts_entry_box.entry_get)
        self.selected_animal_cnt, self.selected_bp_cnt = int(self.animal_cnt_entry_box.entry_get), int(self.no_body_parts_entry_box.entry_get)
        self.bp_name_ebs, self.bp_animal_ebs = [], []

        header_col_1 = SimBALabel(parent=self.bp_table_frm, txt='BODY-PART NAME', font=Formats.FONT_HEADER.value, justify='center', width=24)
        header_col_1.grid(row=0, column=0, sticky=NW)
        if self.selected_animal_cnt > 1:
            header_col_2 = SimBALabel(parent=self.bp_table_frm, txt='ANIMAL ID NUMBER', font=Formats.FONT_HEADER.value, justify='center', width=24)
            header_col_2.grid(row=0, column=1, sticky=NW)

        for i in range(self.selected_bp_cnt * self.selected_animal_cnt):
            bp_name_entry = Entry_Box(parent=self.bp_table_frm, fileDescription=str(i + 1), labelwidth=4, entry_box_width=20, justify='center')
            bp_name_entry.grid(row=i+1, column=0, sticky=NW)
            self.bp_name_ebs.append(bp_name_entry)
            if self.selected_animal_cnt > 1:
                animal_id_entry = Entry_Box(parent=self.bp_table_frm, fileDescription="", labelwidth=4, entry_box_width=20, justify='center', validation='numeric')
                animal_id_entry.grid(row=i+1, column=1, sticky=NW)
                self.bp_animal_ebs.append(animal_id_entry)
        self.bp_table_frm.grid(row=5, sticky=W, column=0)
        self.save_btn.config(state="normal")

    def validate_unique_entries(self, body_part_names: list, animal_ids: list):
        for cnt, bp_name in enumerate(body_part_names):
            check_str(name=f'BODY-PART NAME row {cnt+1}', value=bp_name, allow_blank=False)
            if "," in bp_name: raise InvalidInputError(msg=f'Commas are not allowed in body-part names. A comma was found in body-part name: {bp_name}.', source=self.__class__.__name__)

        for cnt, animal_id in enumerate(animal_ids):
            check_int(name=f'ANIMAL ID NUMBER row {cnt+1}', value=animal_id, min_value=1, max_value=int(self.animal_cnt_entry_box.entry_get), raise_error=True)
        if len(animal_ids) > 0:
            user_entries = []
            for bp_name, animal_id in zip(body_part_names, animal_ids):
                user_entries.append(f"{bp_name}_{animal_id}")
        else:
            user_entries = body_part_names
        duplicates = list(set([x for x in user_entries if user_entries.count(x) > 1]))
        if duplicates:
            print(duplicates)
            raise DuplicationError(msg="SIMBA ERROR: SimBA found duplicate body-part names (see above). Please enter unique body-part names(or body-part/animal ID names , if tracking multiple animals).", source=self.__class__.__name__)
        else:
            pass

    def save_pose_config(self):
        config_name = self.config_name_entry_box.entry_get
        image_path = self.img_path_file_select.file_path
        check_valid_img_path(path=image_path)
        check_str(name="POSE CONFIG NAME", value=config_name.strip(), allow_blank=False)
        bp_lst, animal_id_lst = [], []
        for bp_name_entry in self.bp_name_ebs: bp_lst.append(bp_name_entry.entry_get.strip())
        for animal_id_entry in self.bp_animal_ebs: animal_id_lst.append(animal_id_entry.entry_get.strip())
        self.validate_unique_entries(body_part_names=bp_lst, animal_ids=animal_id_lst)
        animal_id_lst = [int(x) for x in animal_id_lst]
        pose_config_creator = PoseConfigCreator(pose_name=config_name, animal_cnt=int(self.selected_animal_cnt), img_path=image_path, bp_list=bp_lst, animal_id_int_list=animal_id_lst)
        pose_config_creator.launch()
        stdout_success(msg=f'User-defined pose-configuration "{config_name}" created.', source=self.__class__.__name__)
        self.main_frm.winfo_toplevel().destroy()
        self.master.winfo_toplevel().destroy()
        self.project_config_class()
        self.root.destroy()


#CreateUserDefinedPoseConfigurationPopUp()



__author__ = "Simon Nilsson"

import itertools
import os
from datetime import datetime
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

import simba
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_df_field_is_boolean,
                                check_if_dir_exists, check_if_valid_img,
                                check_instance, check_int, check_str,
                                check_that_column_exist, check_valid_boolean,
                                check_valid_dataframe, check_valid_tuple)
from simba.utils.enums import Paths
from simba.utils.errors import FeatureNumberMismatchError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (bgr_to_rgb_tuple,
                                    read_shap_feature_categories_csv,
                                    read_shap_img_paths)

SIMBA_DIR = os.path.dirname(simba.__file__)

def _create_shap_base_plot(baseline_value: int) -> Tuple[np.ndarray, Tuple[int, int], List[Tuple[int, int]]]:

    shap_img_path = os.path.join(SIMBA_DIR, Paths.SIMBA_SHAP_IMG_PATH.value)
    check_if_dir_exists(in_dir=shap_img_path)
    scale_img_paths, category_img_paths = read_shap_img_paths()
    for k, v in category_img_paths.items(): check_file_exist_and_readable(file_path=v)
    img = 255 * np.ones([1680, 1680, 3], dtype=np.uint8)
    baseline_scale_img = cv2.imread(scale_img_paths["baseline_scale"])
    baseline_scale_top_left = (100, 800)
    baseline_scale_bottom_right = (baseline_scale_top_left[0] + baseline_scale_img.shape[0], baseline_scale_top_left[1] + baseline_scale_img.shape[1])
    baseline_scale_middle = ((int(700 + baseline_scale_img.shape[1] / 2)), 85)
    img[baseline_scale_top_left[0] : baseline_scale_bottom_right[0], baseline_scale_top_left[1] : baseline_scale_bottom_right[1]] = baseline_scale_img
    cv2.putText(img, "BASELINE SHAP", baseline_scale_middle, cv2.FONT_HERSHEY_COMPLEX, 1, (0, 0, 0), 2)

    small_arrow_img = cv2.imread(scale_img_paths["small_arrow"])
    small_arrow_top_left = (baseline_scale_bottom_right[0], int(baseline_scale_top_left[1] + (baseline_scale_img.shape[1] / 100) * (baseline_value)))
    small_arrow_bottom_right = (small_arrow_top_left[0] + small_arrow_img.shape[0], small_arrow_top_left[1] + small_arrow_img.shape[1])
    img[small_arrow_top_left[0] : small_arrow_bottom_right[0], small_arrow_top_left[1] : small_arrow_bottom_right[1]] = small_arrow_img

    side_scale_img = cv2.imread(scale_img_paths["side_scale"])
    side_scale_top_left = (small_arrow_bottom_right[0] + 50, baseline_scale_top_left[1] - 50)
    side_scale_y_tick_cords = [(side_scale_top_left[0], side_scale_top_left[1] - 75)]
    for i in range(1, 7):
        side_scale_y_tick_cords.append((int(side_scale_top_left[0] + (side_scale_img.shape[0] / 4) * i), int(side_scale_top_left[1] - 75)))
    arrow_start = (int(small_arrow_top_left[1] + (small_arrow_img.shape[1] / 2)), side_scale_top_left[0])
    for img_cnt, (img_name, img_data) in enumerate(category_img_paths.items()):
        icon_img = cv2.resize(cv2.imread(img_data), None, fx=1.5, fy=1.5, interpolation=cv2.INTER_CUBIC)
        icon_top_left = (side_scale_y_tick_cords[img_cnt][0] - int(icon_img.shape[0] / 2), side_scale_y_tick_cords[img_cnt][1] - 100)
        icon_bottom_right = (icon_top_left[0] + icon_img.shape[0], side_scale_y_tick_cords[img_cnt][1] + icon_img.shape[1] - 100)
        text_location = (int(icon_bottom_right[0] - (icon_bottom_right[0] - icon_top_left[0]) + 100), int(icon_bottom_right[1] - (icon_bottom_right[1] - icon_top_left[1])) - 380)
        cv2.putText(img, str(img_name), (text_location[1], text_location[0]), cv2.FONT_HERSHEY_COMPLEX, 0.75, (0, 0, 0), 1)
        img[icon_top_left[0] : icon_bottom_right[0], icon_top_left[1] : icon_bottom_right[1]] = icon_img

    return img, arrow_start, side_scale_y_tick_cords

def _insert_data_in_base_shap_plot(img: np.ndarray,
                                   arrow_start: Tuple[int, int],
                                   present_df: pd.DataFrame,
                                   side_scale_y_tick_cords: List[Tuple[int, int]]) -> np.ndarray:

    check_if_valid_img(data=img, source=f'{_insert_data_in_base_shap_plot.__name__} img')
    check_valid_tuple(x=arrow_start, min_integer=0)
    check_valid_dataframe(df=present_df, source=f'{_insert_data_in_base_shap_plot.__name__} present_df')
    scale_img_paths, category_img_paths = read_shap_img_paths()
    positive_arrow_colors = [(253, 141, 60), (252, 78, 42), (227, 26, 28), (189, 0, 38), (128, 0, 38)]
    negative_arrow_colors = [(65, 182, 196), (29, 145, 192), (34, 94, 168), (37, 52, 148), (8, 29, 88)]
    positive_arrow_colors = [bgr_to_rgb_tuple(x) for x in positive_arrow_colors]
    negative_arrow_colors = [bgr_to_rgb_tuple(x) for x in negative_arrow_colors]
    ranges_lst = [list(range(0, 20)), list(range(20, 40)), list(range(40, 60)), list(range(60, 80)), list(range(80, 101))]

    baseline_scale_img = cv2.imread(scale_img_paths["baseline_scale"])
    small_arrow_img = cv2.imread(scale_img_paths["small_arrow"])
    color_bar_img = cv2.imread(scale_img_paths["color_bar"])
    baseline_scale_top_left, arrow_end = (100, 800), None

    for row_cnt, (x_name, x_data) in enumerate(present_df.iterrows()):
        x_cat_sum = int(x_data.sum())
        arrow_width = int((baseline_scale_img.shape[1] / 100) * abs(x_cat_sum))
        if x_cat_sum > 0:
            arrow_end = (arrow_start[0] + arrow_width, arrow_start[1])
            arrow_middle = int(((arrow_end[1] - arrow_start[1]) / 2) + arrow_start[1] - 7)
            for bracket_num, bracket in enumerate(ranges_lst):
                if abs(x_cat_sum) in bracket:
                    color = positive_arrow_colors[bracket_num]
                    cv2.arrowedLine(img, arrow_start, arrow_end, color, 5, tipLength=0.1)
                    cv2.putText(img, f"+{x_cat_sum}%", (arrow_end[0] - 7, arrow_middle - 15), cv2.FONT_HERSHEY_COMPLEX, 1,  color, 2,)
        else:
            arrow_end = (arrow_start[0] - arrow_width, arrow_start[1])
            arrow_middle = int(((arrow_start[1] - arrow_end[1]) / 2) + arrow_end[1] - 7)
            for bracket_num, bracket in enumerate(ranges_lst):
                if abs(x_cat_sum) in bracket:
                    color = negative_arrow_colors[bracket_num]
                    cv2.arrowedLine(img, arrow_start, arrow_end, color, 5, tipLength=0.1)
                    cv2.putText(img, f"-{abs(x_cat_sum)}%", (arrow_end[0] - 7, arrow_middle - 15), cv2.FONT_HERSHEY_COMPLEX, 1, color, 2)

        if row_cnt != (len(present_df) - 1):
            arrow_start = (arrow_end[0], side_scale_y_tick_cords[row_cnt + 1][0])

    small_arrow_top_left = (int(arrow_end[1]) + 20, int(arrow_end[0] - small_arrow_img.shape[1] / 2))
    small_arrow_bottom_right = (small_arrow_top_left[0] + small_arrow_img.shape[0], small_arrow_top_left[1] + small_arrow_img.shape[1])
    img[small_arrow_top_left[0] : small_arrow_bottom_right[0], small_arrow_top_left[1] : small_arrow_bottom_right[1]] = small_arrow_img
    color_bar_top_left = (arrow_end[1] + small_arrow_img.shape[0] + 25, baseline_scale_top_left[1])
    color_bar_bottom_right = (color_bar_top_left[0] + color_bar_img.shape[0], color_bar_top_left[1] + color_bar_img.shape[1])
    img[color_bar_top_left[0] : color_bar_bottom_right[0], color_bar_top_left[1] : color_bar_bottom_right[1]] = color_bar_img
    color_bar_middle = ((int(580 + baseline_scale_img.shape[1] / 2)), color_bar_bottom_right[0] + 50)
    cv2.putText(img, "CLASSIFICATION PROBABILITY", color_bar_middle, cv2.FONT_HERSHEY_COMPLEX, 1, (0, 0, 0), 2)
    return img

class ShapAggregateStatisticsCalculator():
    """
    Calculate aggregate (binned) SHAP value statistics where individual bins represent reaulated features.
    and create line chart visualizations reprsenting aggregations of behavior-present SHAP values.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/SHAP.md#step-3-interpreting-the-shap-value-ouput-generated-by-simba>`__.
       `Example output <https://github.com/sgoldenlab/simba/blob/master/images/example_shap_graph.png>`__.

    .. image:: _static/img/shap.png
       :width: 600
       :align: center

    :param pd.DataFrame shap_df: Data with framewise SHAP values.
    :param str classifier_name: Name of classifier (e.g., Attack).
    :param pd.DataFrame shap_df: Dataframe with non-aggregated SHAP values where rows represent frames and columns represent features.
    :param float shap_baseline_value: SHAP expected value (computed by ``simba.train_model_functions.create_shap_log``).
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where to store the results. If None, then return the results instead of saving it.
    :param Optional[Any] filename_suffix: Optional suffix to add to the shap output filenames. Useful for gridsearches and multiple shap data output files are to-be stored in the same `save_dir`.
    :param bool plot: If True, creates a visualization of the aggregate SHAP values. Default True.

    :example:
    >>> shap_df = pd.read_csv('/Users/simon/Desktop/envs/simba/simba/tests/data/sample_data/shap_test.csv', index_col=0)
    >>> test = ShapAggregateStatisticsCalculator(classifier_name='target',
    >>>                                          shap_df=shap_df,
    >>>                                          shap_baseline_value=40,
    >>>                                          save_dir=None) #'/Users/simon/Desktop/feltz'
    >>> dfs, img = test.run()
    """

    def __init__(self,
                 shap_df: pd.DataFrame,
                 classifier_name: str,
                 shap_baseline_value: int,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 filename_suffix: Optional[int] = None,
                 plot: bool = True):

        check_instance(source=f"{self.__class__.__name__} shap_df", instance=shap_df, accepted_types=(pd.DataFrame,))
        check_str(name=f"{self.__class__.__name__} classifier_name", value=classifier_name)
        check_that_column_exist(df=shap_df, column_name=classifier_name, file_name="shap dataframe")
        check_if_df_field_is_boolean(df=shap_df, field=classifier_name)
        check_int(name=f"{self.__class__.__name__} shap_baseline_value", value=shap_baseline_value, max_value=100, min_value=0)
        check_valid_boolean(value=[plot], source=f'{self.__class__.__name__} plot')
        self.clf_name, self.shap_df, self.shap_baseline_value, self.plot = classifier_name, shap_df, shap_baseline_value, plot
        self.save_path, self.datetime = None, datetime.now().strftime("%Y%m%d%H%M%S")
        self.present_save_path, self.absent_save_path, img_save_path = None, None, None
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir)
            if filename_suffix is None:
                self.img_save_path = os.path.join(save_dir, f"SHAP_summary_line_graph_{self.clf_name}.png")
                self.df_save_paths = {'PRESENT': os.path.join(save_dir, f"SHAP_summary_{self.clf_name}_PRESENT.csv"),
                                      'ABSENT': os.path.join(save_dir, f"SHAP_summary_{self.clf_name}_ABSENT.csv")}
            else:
                self.img_save_path = os.path.join(save_dir, f"SHAP_summary_line_graph_{self.clf_name}_{filename_suffix}.png")
                self.df_save_paths = {'PRESENT': os.path.join(save_dir, f"SHAP_summary_{self.clf_name}_PRESENT_{filename_suffix}.csv"),
                                      'ABSENT': os.path.join(save_dir, f"SHAP_summary_{self.clf_name}_ABSENT_{filename_suffix}.csv")}
        self.feature_categories_df, self.x_names, self.unique_feature_category_names, self.unique_time_bin_names = read_shap_feature_categories_csv()
        self.save_dir = save_dir
        v = [x for x in list(shap_df.columns) if x in self.x_names]
        if len(v) == 0:
            raise FeatureNumberMismatchError('No feature names in the shap dataframe is defined in the shap aggregation classes', source=self.__class__.__name__)

    def run(self):
        timer = SimbaTimer(start=True)
        self.results, self.img = {}, None
        for state, clf_state_name in zip(range(2), ["ABSENT", "PRESENT"]):
            state_result = {}
            shap_df = self.shap_df[self.shap_df[self.clf_name] == state]
            for x_category, x_time_bin in itertools.product(self.unique_feature_category_names, self.unique_time_bin_names):
                if x_category not in state_result.keys():
                    state_result[x_category] = {}
                feature_names_sliced = list(self.feature_categories_df.loc[:, (x_category, x_time_bin)])
                feature_names_sliced = [x for x in feature_names_sliced if str(x) != "nan" and x in feature_names_sliced]
                feature_names_sliced = [x for x in feature_names_sliced if x in shap_df.columns]
                state_result[x_category][x_time_bin] = round(shap_df[feature_names_sliced].sum(axis=1).mean() * 100, 6)
            results = pd.DataFrame(columns=state_result[list(state_result.keys())[0]], index=state_result.keys())
            for row_name, time_bins in state_result.items():
                for column_name, value in time_bins.items():
                    results.loc[row_name, column_name] = value
            results.reindex(sorted(results.columns, reverse=True), axis=1)
            self.results[clf_state_name] = results
            if self.save_dir is not None:
                self.results[clf_state_name].to_csv(self.df_save_paths[clf_state_name])

        if self.plot:
            base_plot, arrow_start, side_scale_y_tick_cords = _create_shap_base_plot(baseline_value=self.shap_baseline_value)
            self.img = _insert_data_in_base_shap_plot(img=base_plot, arrow_start=arrow_start, present_df=self.results['PRESENT'], side_scale_y_tick_cords=side_scale_y_tick_cords)
            if self.save_dir is not None:
                cv2.imwrite(self.img_save_path, self.img)

        timer.stop_timer()
        if self.save_dir is not None:
            stdout_success(msg=f"SHAP summary graph saved at {self.img_save_path}", elapsed_time=timer.elapsed_time_str)
        else:
            return (self.results, self.img)


# shap_df = pd.read_csv('/Users/simon/Desktop/envs/simba/simba/tests/data/sample_data/shap_test.csv', index_col=0)
# test = ShapAggregateStatisticsCalculator(classifier_name='target',
#                                          shap_df=shap_df,
#                                          shap_baseline_value=40,
#                                          save_dir=None) #'/Users/simon/Desktop/feltz'
# dfs, img = test.run()

# shap_df = pd.read_csv('/Users/simon/Desktop/envs/simba/simba/tests/data/sample_data/shap_test.csv', index_col=0)
# test = ShapAggregateStatisticsVisualizer(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini',
#                                          classifier_name='target',
#                                          shap_df=shap_df,
#                                          shap_baseline_value=40,
#                                          save_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/shap')
#



import os
from typing import List, Optional, Tuple, Union

import numpy as np

from simba.mixins.geometry_mixin import GeometryMixin
from simba.plotting.geometry_plotter import GeometryPlotter
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_if_valid_rgb_tuple,
                                check_instance, check_int,
                                check_valid_dataframe)
from simba.utils.enums import Formats
from simba.utils.errors import CountError, NoFilesFoundError
from simba.utils.lookups import get_current_time
from simba.utils.read_write import (find_all_videos_in_directory,
                                    find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df)

VERTICE = 'VERTICE'
FRAME = 'FRAME'
NAME = 'NAME'


class SamVisualizer():
    """
    :example:
    >>> r = SamVisualizer(data_path=r"C:\troubleshooting\sam_results\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7_clipped_3.csv", video_path=r"D:\platea\platea_videos\videos\clipped\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7_clipped_3.mp4", save_dir='D:\cvat_annotations\sam_videos', color=[(255, 255, 1)])
    >>> r.run()
    """

    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 core_cnt: int = -1,
                 color: Optional[Union[Tuple[int, int, int], List[Tuple[int, int, int]]]] = (178, 102, 255),
                 shape_opacity: float = 0.5,
                 bg_opacity: float = 1.0):

        if os.path.isdir(data_path):
            self.data_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=['.csv'],
                                                                   raise_error=True, as_dict=True)
        elif os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.data_paths = {get_fn_ext(filepath=data_path)[1]: data_path}
        else:
            raise NoFilesFoundError(msg=f'{data_path} is not a valid file path or directory.',
                                    source=self.__class__.__name__)
        if os.path.isdir(video_path):
            self.video_paths = find_all_videos_in_directory(directory=video_path, raise_error=True, as_dict=True)
        elif os.path.isfile(video_path):
            check_file_exist_and_readable(file_path=video_path)
            self.video_paths = {get_fn_ext(filepath=video_path)[1]: video_path}
        else:
            raise NoFilesFoundError(msg=f'{video_path} is not a valid directory of file path.')
        for k, v in self.data_paths.items():
            if k not in self.video_paths.keys():
                raise NoFilesFoundError(
                    msg=f'Could not find a video for SAM tracking data file {k} in the {video_path} directory',
                    source=self.__class__.__name__)
        check_if_dir_exists(in_dir=save_dir, create_if_not_exist=True)
        check_int(name='core count', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt
        check_float(name='shape_opacity', value=shape_opacity, max_value=1.0, min_value=0.1, raise_error=True)
        check_float(name='bg_opacity', value=bg_opacity, max_value=1.0, min_value=0.1, raise_error=True)
        check_instance(source=f'{self.__class__.__name__} color', accepted_types=(list, tuple,), instance=color)
        if isinstance(color, (tuple,)):
            check_if_valid_rgb_tuple(data=color, raise_error=True, source=self.__class__.__name__)
            color = [color]
        else:
            for c in color:
                check_if_valid_rgb_tuple(data=c)
        self.bg_opacity, self.shape_opacity, self.colors = bg_opacity, shape_opacity, color
        self.save_dir = save_dir

    def run(self):
        print(
f'Running SAM visualization for {len(list(self.data_paths.keys()))} video(s) (start time: {get_current_time()})...')
        for file_name, file_path in self.data_paths.items():
            print(f'Running SAM visualization for video {file_name}...')
            geometries = []
            video_path = self.video_paths[file_name]
            df = read_df(file_path=file_path, file_type='csv')
            animal_names = df[NAME].unique()
            animal_names = [x for x in animal_names if x != -1]
            if len(animal_names) != len(self.colors):
                raise CountError(msg=f'{len(animal_names)} animal names ({animal_names}) found in data {file_path} but {len(self.colors)} color(s) passed: {self.colors}', source=self.__class__.__name__)
            # check_video_and_data_frm_count_align(video=video_path, data=df, name=file_name, raise_error=True)
            check_valid_dataframe(df=df, source='', valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=[FRAME, NAME])
            for animal_name in animal_names:
                animal_df = df[df[NAME] == animal_name].reset_index(drop=True)
                vertice_cols = [x for x in animal_df.columns if VERTICE in x]
                animal_vertice_arr = animal_df[vertice_cols].values.reshape(len(animal_df), int(len(vertice_cols) / 2), 2).astype(np.int32)
                geometries.append(GeometryMixin().bodyparts_to_polygon(data=animal_vertice_arr, simplify_tolerance=2, convex_hull=False))

            plotter = GeometryPlotter(geometries=geometries,
                                      video_name=video_path,
                                      save_dir=self.save_dir,
                                      core_cnt=self.core_cnt,
                                      palette=None,
                                      colors=self.colors,
                                      bg_opacity=self.bg_opacity,
                                      shape_opacity=self.shape_opacity,
                                      circle_size=None)
            plotter.run()
            #


r = SamVisualizer(
    data_path=r"D:\cvat_annotations\sam_yolo_data\s34-drinking.csv",
    video_path=r"D:\cvat_annotations\videos\mp4_20250624155703\s34-drinking.mp4",
    save_dir='D:\cvat_annotations\sam_videos', color=[(255, 255, 1)])
r.run()


__author__ = "Simon Nilsson"

import os
import threading
from collections import defaultdict
from tkinter import *
from typing import Union

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.distance_plotter import DistancePlotterSingleCore
from simba.plotting.distance_plotter_mp import DistancePlotterMultiCore
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, SimbaButton, SimbaCheckbox,
                                        SimBADropDown)
from simba.utils.checks import check_if_filepath_list_is_empty, check_int
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import DuplicationError
from simba.utils.read_write import get_file_name_info_in_directory

AUTO = 'AUTO'

class DistancePlotterPopUp(PopUpMixin, ConfigReader):

    """
    :example:
    >>> _ = DistancePlotterPopUp(config_path=r'C:\troubleshooting\RAT_NOR\project_folder\project_config.ini')

    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.max_y_lst = list(range(10, 1010, 10))
        self.max_y_lst.insert(0, AUTO)
        font_size_options = list(range(1, 33))
        opacity_options = list(np.round(np.arange(0.0, 1.1, 0.1), 1))
        self.number_of_distances_options = list(range(1, len(self.body_parts_lst) * 2))

        self.files_found_dict = get_file_name_info_in_directory(directory=self.outlier_corrected_dir, file_type=self.file_type)
        check_if_filepath_list_is_empty(filepaths=list(self.files_found_dict.keys()), error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/outlier_corrected_movement_location directory. ",)

        PopUpMixin.__init__(self, title="CREATE DISTANCE PLOTS", icon='distance')

        self.style_settings_frm = CreateLabelFrameWithIcon( parent=self.main_frm, header="STYLE SETTINGS", icon_name='style', icon_link=Links.DISTANCE_PLOTS.value,)
        self.resolution_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.resolutions, label='RESOLUTION: ', label_width=25, dropdown_width=20, value=self.resolutions[1])
        self.font_size_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=font_size_options, label='FONT SIZE: ', label_width=25, dropdown_width=20, value=8)
        self.line_width_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=font_size_options, label='LINE WIDTH: ', label_width=25, dropdown_width=20, value=6)
        self.opacity_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=opacity_options, label='LINE OPACITY: ', label_width=25, dropdown_width=20, value=0.5)
        self.max_y_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.max_y_lst, label='MAX Y-AXIS: ', label_width=25, dropdown_width=20, value=AUTO)

        self.distances_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="CHOOSE DISTANCES", icon_name='distance', icon_link=Links.DISTANCE_PLOTS.value)
        self.number_of_distances_dropdown = SimBADropDown(parent=self.distances_frm, dropdown_options=self.number_of_distances_options, label='# DISTANCES: ', label_width=25, dropdown_width=20, value=1, command=self.__populate_distances_menu)
        self.__populate_distances_menu(1)

        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VISUALIZATION SETTING", icon_name='eye', icon_link=Links.DISTANCE_PLOTS.value)

        distance_frames_cb, self.distance_frames_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE FRAMES', txt_img='frames', val=False)
        distance_videos_cb, self.distance_videos_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE VIDEOS', txt_img='video', val=False)
        distance_final_img_cb, self.distance_final_img_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE LAST FRAME', txt_img='finish', val=True)
        self.multiprocess_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, self.cpu_cnt)), label='CPU CORE COUNT: ', label_width=25, dropdown_width=20, value=int(self.cpu_cnt/2))

        self.run_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="RUN", icon_name='run', icon_link=Links.DISTANCE_PLOTS.value)
        self.run_single_video_frm = LabelFrame(self.run_frm, text="SINGLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black")

        self.run_single_video_btn = SimbaButton(parent=self.run_single_video_frm, txt="Create single video", txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.__create_distance_plots, cmd_kwargs={'multiple_videos': False})
        self.single_video_dropdown = DropDownMenu( self.run_single_video_frm, "Video:", list(self.files_found_dict.keys()), "12")
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])

        self.run_multiple_videos = LabelFrame(self.run_frm, text="MULTIPLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black",)
        self.run_multiple_video_btn = SimbaButton(parent=self.run_multiple_videos, txt="Create multiple videos ({} video(s) found)".format(str(len(list(self.files_found_dict.keys())))), txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.__create_distance_plots, cmd_kwargs={'multiple_videos': True})

        self.style_settings_frm.grid(row=0, sticky=NW)
        self.resolution_dropdown.grid(row=0, sticky=NW)
        self.font_size_dropdown.grid(row=1, sticky=NW)
        self.line_width_dropdown.grid(row=2, sticky=NW)
        self.opacity_dropdown.grid(row=3, sticky=NW)
        self.max_y_dropdown.grid(row=4, sticky=NW)

        self.distances_frm.grid(row=1, sticky=NW)
        self.number_of_distances_dropdown.grid(row=0, sticky=NW)

        self.settings_frm.grid(row=2, sticky=NW)
        self.multiprocess_dropdown.grid(row=0, column=0, sticky=NW)
        distance_frames_cb.grid(row=1, sticky=NW)
        distance_videos_cb.grid(row=2, sticky=NW)
        distance_final_img_cb.grid(row=3, sticky=NW)

        self.run_frm.grid(row=3, sticky=NW)
        self.run_single_video_frm.grid(row=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=1, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)

        self.main_frm.mainloop()

    def __populate_distances_menu(self, choice):
        if hasattr(self, "bp_1"):
            for k, v in self.bp_1.items():
                self.bp_1[k].destroy()
                self.bp_2[k].destroy()
                self.distance_clrs[k].destroy()

        self.bp_1, self.bp_2, self.distance_clrs = {}, {}, {}
        for distance_cnt in range(int(self.number_of_distances_dropdown.getChoices())):
            self.bp_1[distance_cnt] = SimBADropDown(parent=self.distances_frm, dropdown_options=self.body_parts_lst, label=f'DISTANCE {distance_cnt + 1}: ', label_width=25, dropdown_width=20, value=self.body_parts_lst[distance_cnt])
            self.bp_1[distance_cnt].grid(row=distance_cnt + 1, column=0, sticky=NW)
            self.bp_2[distance_cnt] = SimBADropDown(parent=self.distances_frm, dropdown_options=self.body_parts_lst, label='', label_width=0, dropdown_width=20, value=self.body_parts_lst[distance_cnt])
            self.bp_2[distance_cnt].grid(row=distance_cnt + 1, column=1, sticky=NW)
            self.distance_clrs[distance_cnt] = SimBADropDown(parent=self.distances_frm, dropdown_options=list(self.colors_dict.keys()), label='', label_width=0, dropdown_width=20, value=list(self.colors_dict.keys())[distance_cnt])
            self.distance_clrs[distance_cnt].grid(row=distance_cnt + 1, column=3, sticky=NW)

    def __create_distance_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [self.files_found_dict[self.single_video_dropdown.getChoices()]]

        line_attr = defaultdict(list)
        for attr in (self.bp_1, self.bp_2, self.distance_clrs):
            for key, value in attr.items():
                line_attr[key].append(value.getChoices())
        line_attr = list(line_attr.values())

        for cnt, v in enumerate(line_attr):
            if v[0] == v[1]:
                raise DuplicationError(msg=f"DISTANCE LINE {cnt+1} ERROR: The two body-parts cannot be the same body-part.",source=self.__class__.__name__,)

        width = int(self.resolution_dropdown.getChoices().split("×")[0])
        height = int(self.resolution_dropdown.getChoices().split("×")[1])
        font_size = int(self.font_size_dropdown.get_value())
        line_width = int(self.line_width_dropdown.get_value())
        opacity = float(self.opacity_dropdown.getChoices())
        core_cnt = int(self.multiprocess_dropdown.get_value())
        max_y = -1 if self.max_y_dropdown.getChoices() == AUTO else int(self.max_y_dropdown.getChoices())

        style_attr = {"width": width,
                      "height": height,
                      "line width": line_width,
                      "font size": font_size,
                      "opacity": opacity,
                      "y_max": max_y}

        if core_cnt == 1:
            distance_plotter = DistancePlotterSingleCore(config_path=self.config_path,
                                                         frame_setting=self.distance_frames_var.get(),
                                                         video_setting=self.distance_videos_var.get(),
                                                         final_img=self.distance_final_img_var.get(),
                                                         style_attr=style_attr,
                                                         data_paths=data_paths,
                                                         line_attr=line_attr)
        else:
            distance_plotter = DistancePlotterMultiCore(config_path=self.config_path,
                                                        frame_setting=self.distance_frames_var.get(),
                                                        video_setting=self.distance_videos_var.get(),
                                                        final_img=self.distance_final_img_var.get(),
                                                        style_attr=style_attr,
                                                        data_paths=data_paths,
                                                        line_attr=line_attr,
                                                        core_cnt=core_cnt)

        threading.Thread(distance_plotter.run()).start()



#_ = DistancePlotterPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")


# _ = DistancePlotterPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = DistancePlotterPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')



import os
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd

from simba.utils.checks import (check_if_dir_exists, check_if_valid_rgb_tuple,
                                check_int, check_str, check_valid_boolean,
                                check_valid_dataframe, check_valid_tuple)
from simba.utils.data import egocentrically_align_pose_numba
from simba.utils.enums import Formats, Options
from simba.utils.errors import InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (bgr_to_rgb_tuple, find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df, write_df)
from simba.video_processors.egocentric_video_rotator import \
    EgocentricVideoRotator


class EgocentricalAligner():

    """
    Aligns and rotates movement data and associated video frames based on specified anchor points to produce an egocentric view of the subject. The class aligns frames around a selected anchor point, optionally rotating the subject to a consistent direction and saving the output video.

    .. video:: _static/img/EgocentricalAligner.webm
       :width: 800
       :autoplay:
       :loop:

    .. video:: _static/img/EgocentricalAligner_2.webm
       :width: 800
       :autoplay:
       :loop:

    .. seealso::
       To produce rotation vectors, uses :func:`~simba.utils.data.egocentrically_align_pose_numba` or :func:`~simba.utils.data.egocentrically_align_pose`.
       To rotate video only, see :func:`~simba.video_processors.egocentric_video_rotator.EgocentricVideoRotator`


    :param Union[str, os.PathLike] config_path: Path to the configuration file.
    :param Union[str, os.PathLike] save_dir: Directory where the processed output will be saved.
    :param Optional[Union[str, os.PathLike]] data_dir: Directory containing CSV files with movement data.
    :param Optional[str] anchor_1: Primary anchor point (e.g., 'tail_base') around which the alignment centers.
    :param Optional[str] anchor_2: Secondary anchor point (e.g., 'nose') defining the alignment direction.
    :param int direction: Target angle, in degrees, for alignment; e.g., `0` aligns east
    :param Optional[Tuple[int, int]] anchor_location: Pixel location in the output where `anchor_1`  should appear; default is `(250, 250)`.
    :param Tuple[int, int, int] fill_clr: If rotating the videos, the color of the additional pixels.
    :param Optional[bool] rotate_video: Whether to rotate the video to align with the specified direction.
    :param Optional[int] core_cnt: Number of CPU cores to use for video rotation; `-1` uses all available cores.

    :example:
     >>> aligner = EgocentricalAligner(rotate_video=True, anchor_1='tail_base', anchor_2='nose', data_dir=r"/data_dir", videos_dir=r'/videos_dir', save_dir=r"/save_dir", video_info=r"C:\troubleshooting\mitra\project_folder\logs\video_info.csv", direction=0, anchor_location=(250, 250), fill_clr=(0, 0, 0))
     >>> aligner.run()
    """

    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 anchor_1: str = 'tail_base',
                 anchor_2: str = 'nose',
                 direction: int = 0,
                 core_cnt: int = -1,
                 fill_clr: Tuple[int, int, int] = (250, 250, 255),
                 verbose: bool = True,
                 gpu: bool = False,
                 videos_dir: Optional[Union[str, os.PathLike]] = None,
                 anchor_location: Optional[Union[Tuple[int, int], str]] = (250, 250)):

        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'])
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir')
        check_str(name=f'{self.__class__.__name__} anchor_1', value=anchor_1, allow_blank=False)
        check_str(name=f'{self.__class__.__name__} anchor_2', value=anchor_2, allow_blank=False)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, max_value=find_core_cnt()[0], unaccepted_vals=[0])
        if core_cnt == -1: self.core_cnt = find_core_cnt()[0]
        check_int(name=f'{self.__class__.__name__} direction', value=direction, min_value=0, max_value=360)
        if isinstance(anchor_location, tuple):
            check_valid_tuple(x=anchor_location, source=f'{self.__class__.__name__} anchor_location', accepted_lengths=(2,), valid_dtypes=(int,))
            for i in anchor_location:
                check_int(name=f'{self.__class__.__name__} anchor_location', value=i, min_value=1)
        if videos_dir is None and anchor_location is None:
            raise InvalidInputError(msg='If anchor_location is None, please pass video_dir')
        check_valid_boolean(value=[gpu], source=f'{self.__class__.__name__} gpu')
        if videos_dir is not None:
            check_if_valid_rgb_tuple(data=fill_clr)
            fill_clr = bgr_to_rgb_tuple(value=fill_clr)
            check_if_dir_exists(in_dir=videos_dir, source=f'{self.__class__.__name__} videos_dir')
            self.video_paths = find_files_of_filetypes_in_directory(directory=videos_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value)
            for file_path in self.data_paths:
                find_video_of_file(video_dir=videos_dir, filename=get_fn_ext(file_path)[1], raise_error=True)
        self.anchor_1_cols = [f'{anchor_1}_x'.lower(), f'{anchor_1}_y'.lower()]
        self.anchor_2_cols = [f'{anchor_2}_x'.lower(), f'{anchor_2}_y'.lower()]
        self.anchor_1, self.anchor_2, self.videos_dir = anchor_1.lower(), anchor_2.lower(), videos_dir
        self.save_dir, self.verbose, self.gpu = save_dir, verbose, gpu
        self.anchor_location, self.direction, self.fill_clr = anchor_location, direction, fill_clr
        self.rotate_video = [True if videos_dir is not None else False]

    def run(self):
        timer = SimbaTimer(start=True)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            if self.verbose:
                print(f'Rotating data video {self.video_name}... ({file_cnt+1}/{len(self.data_paths)})')
            if self.anchor_location is None:
                video_path = find_video_of_file(video_dir=self.videos_dir, filename=self.video_name, raise_error=False)
                video_meta_data = get_video_meta_data(video_path=video_path)
                self.anchor_location = (int(video_meta_data['width']/2), int(video_meta_data['height']/2))
            save_path = os.path.join(self.save_dir, f'{self.video_name}.{Formats.CSV.value}')
            df = read_df(file_path=file_path, file_type=Formats.CSV.value)
            original_cols, self.file_path = list(df.columns), file_path
            df.columns = [x.lower() for x in list(df.columns)]
            check_valid_dataframe(df=df, source=self.__class__.__name__, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.anchor_1_cols + self.anchor_2_cols)
            bp_cols = [x for x in df.columns if not x.endswith('_p')]
            body_parts_lst = []
            _= [body_parts_lst.append(x[:-2]) for x in bp_cols if x[:-2] not in body_parts_lst]
            anchor_1_idx, anchor_2_idx = body_parts_lst.index(self.anchor_1), body_parts_lst.index(self.anchor_2)
            data_arr = df[bp_cols].values.reshape(len(df), len(body_parts_lst), 2).astype(np.int32)
            results_arr, self.centers, self.rotation_vectors = egocentrically_align_pose_numba(data=data_arr, anchor_1_idx=anchor_1_idx, anchor_2_idx=anchor_2_idx, direction=self.direction, anchor_location=np.array(self.anchor_location).astype(np.int64))
            results_arr = results_arr.reshape(len(df), len(bp_cols))
            self.out_df = pd.DataFrame(results_arr, columns=bp_cols)
            df.update(self.out_df)
            df.columns = original_cols
            write_df(df=df, file_type=Formats.CSV.value, save_path=save_path)
            video_timer.stop_timer()
            if self.verbose:
                print(f'{self.video_name} complete, saved at {save_path} (elapsed time: {video_timer.elapsed_time_str}s)')
            if self.rotate_video:
                if self.verbose:
                    print(f'Rotating video {self.video_name}...')
                video_path = find_video_of_file(video_dir=self.videos_dir, filename=self.video_name, raise_error=False)
                save_path = os.path.join(self.save_dir, f'{self.video_name}.mp4')
                video_rotator = EgocentricVideoRotator(video_path=video_path,
                                                       centers=self.centers,
                                                       rotation_vectors=self.rotation_vectors,
                                                       anchor_location=self.anchor_location,
                                                       verbose=self.verbose,
                                                       gpu=self.gpu,
                                                       fill_clr=self.fill_clr,
                                                       core_cnt=self.core_cnt,
                                                       save_path=save_path)
                video_rotator.run()
            if self.verbose:
                print(f'Rotated data for video {self.video_name} ({file_cnt+1}/{len(self.data_paths)}) saved in {self.save_dir}.')
        timer.stop_timer()
        stdout_success(msg=f'Egocentrically aligned data for {len(self.data_paths)} files saved in {self.save_dir}', elapsed_time=timer.elapsed_time_str)


# if __name__ == "__main__":
#     aligner = EgocentricalAligner(anchor_1='butt/proximal tail',
#                                   anchor_2='snout',
#                                   data_dir=r'C:\troubleshooting\open_field_below\project_folder\csv\outlier_corrected_movement_location',
#                                   videos_dir=r'C:\troubleshooting\open_field_below\project_folder\videos',
#                                   save_dir=r"C:\troubleshooting\open_field_below\project_folder\videos\rotated",
#                                   direction=0,
#                                   gpu=True,
#                                   anchor_location=(600, 300),
#                                   fill_clr=(128,128,128))
#     aligner.run()

#     aligner = EgocentricalAligner(anchor_1='tail_base',
#                                   anchor_2='nose',
#                                   data_dir=r'C:\Users\sroni\OneDrive\Desktop\rotate_ex\data',
#                                   videos_dir=r'C:\Users\sroni\OneDrive\Desktop\rotate_ex\videos',
#                                   save_dir=r"C:\Users\sroni\OneDrive\Desktop\rotate_ex\rotated",
#                                   direction=0,
#                                   gpu=True,
#                                   anchor_location=(250, 250),
#                                   fill_clr=(0, 0, 0))
#     aligner.run()

    # aligner = EgocentricalAligner(rotate_video=True,
    #                               anchor_1='tail_base',
    #                               anchor_2='nose',
    #                               data_dir=r'C:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location',
    #                               videos_dir=r'C:\troubleshooting\mitra\project_folder\videos',
    #                               save_dir=r"C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated",
    #                               video_info=r"C:\troubleshooting\mitra\project_folder\logs\video_info.csv")
    # aligner.run()





__author__ = "Simon Nilsson"

import os.path
import platform
import threading
import tkinter
import tkinter as tk
import webbrowser
from copy import copy
from tkinter import *
from tkinter.filedialog import askdirectory, askopenfilename
from tkinter.ttk import Combobox
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import PIL.Image
from PIL import ImageTk

from simba.utils.enums import Defaults, Formats
from simba.utils.lookups import get_icons_paths
from simba.utils.read_write import get_fn_ext

MENU_ICONS = get_icons_paths()


def onMousewheel(event, canvas):
    try:
        scrollSpeed = event.delta
        if platform.system() == "Darwin":
            scrollSpeed = event.delta
        elif platform.system() == "Windows":
            scrollSpeed = int(event.delta / 120)
        canvas.yview_scroll(-1 * (scrollSpeed), "units")
    except:
        pass


def bindToMousewheel(event, canvas):
    canvas.bind_all("<MouseWheel>", lambda event: onMousewheel(event, canvas))


def unbindToMousewheel(event, canvas):
    canvas.unbind_all("<MouseWheel>")


def onFrameConfigure(canvas):
    """Reset the scroll region to encompass the inner frame"""
    canvas.configure(scrollregion=canvas.bbox("all"))


def on_mouse_scroll(event, canvas):
    if event.delta:
        canvas.yview_scroll(-1 * (event.delta // 120), "units")
    elif event.num == 4:
        canvas.yview_scroll(-1, "units")
    elif event.num == 5:
        canvas.yview_scroll(1, "units")

def hxtScrollbar(master):
    canvas = tk.Canvas(master, borderwidth=0, bg="#f0f0f0", relief="flat")
    frame = tk.Frame(canvas, bg="#f0f0f0", bd=0)

    vsb = tk.Scrollbar(master, orient="vertical", command=canvas.yview, bg="#cccccc", width=20)
    hsb = tk.Scrollbar(master, orient="horizontal", command=canvas.xview, bg="#cccccc")

    canvas.configure(yscrollcommand=vsb.set, xscrollcommand=hsb.set)

    canvas.pack(side="left", fill="both", expand=True, padx=0, pady=0)
    vsb.pack(side="right", fill="y", padx=(0, 0))
    hsb.pack(side="bottom", fill="x", pady=(0, 0))

    canvas.create_window((0, 0), window=frame, anchor="nw")

    def on_frame_configure(event):
        canvas.configure(scrollregion=canvas.bbox("all"))

    frame.bind("<Configure>", on_frame_configure)

    def _on_mousewheel(event):
        canvas.yview_scroll(int(-1 * (event.delta / 120)), "units")

    def _on_key(event):
        if event.keysym == "Up":
            canvas.yview_scroll(-1, "units")
        elif event.keysym == "Down":
            canvas.yview_scroll(1, "units")
        elif event.keysym == "Page_Up":
            canvas.yview_scroll(-1, "pages")
        elif event.keysym == "Page_Down":
            canvas.yview_scroll(1, "pages")

    # Focus & bindings when mouse enters
    def _on_enter(_):
        frame.focus_set()
        frame.bind_all("<MouseWheel>", _on_mousewheel)
        frame.bind_all("<Key>", _on_key)

    # Clean up when mouse leaves
    # def _on_leave(_):
    #     frame.unbind_all("<MouseWheel>")
    #     frame.unbind_all("<Key>")

    frame.bind("<Enter>", _on_enter)
    #frame.bind("<Leave>", _on_leave)

    return frame

def form_validator_is_numeric(inStr, acttyp):
    if acttyp == "1":  # insert
        if not inStr.isdigit():
            return False
    return True


class DropDownMenu(Frame):
    def __init__(self,
                 parent=None,
                 dropdownLabel="",
                 choice_dict=None,
                 labelwidth="",
                 com=None,
                 val: Optional[Any] = None,
                 **kw):

        Frame.__init__(self, master=parent, **kw)
        self.dropdownvar = StringVar()
        self.lblName = Label(self, text=dropdownLabel, width=labelwidth, anchor=W, font=Formats.FONT_REGULAR.value)
        self.lblName.grid(row=0, column=0)
        self.choices = choice_dict
        self.popupMenu = OptionMenu(self, self.dropdownvar, *self.choices, command=com)
        self.popupMenu.grid(row=0, column=1)
        if val is not None:
            self.setChoices(val)

    def getChoices(self):
        return self.dropdownvar.get()

    def setChoices(self, choice):
        self.dropdownvar.set(choice)

    def enable(self):
        self.popupMenu.configure(state="normal")

    def disable(self):
        self.popupMenu.configure(state="disable")


class Entry_Box(Frame):
    def __init__(self,
                 parent=None,
                 fileDescription="",
                 labelwidth="",
                 label_bg_clr: Optional[str] = None,
                 status=None,
                 validation=None,
                 entry_box_width=None,
                 value: Optional[Any] = None,
                 label_font: tuple = Formats.FONT_REGULAR.value,
                 entry_font: tuple = Formats.FONT_REGULAR.value,
                 justify: str = 'left',
                 **kw):

        super(Entry_Box, self).__init__(master=parent)
        self.validation_methods = {"numeric": (self.register(form_validator_is_numeric), "%P", "%d")}
        self.status = status if status is not None else NORMAL
        self.labelname = fileDescription
        Frame.__init__(self, master=parent, **kw)
        self.filePath = StringVar()
        self.lblName = Label(self, text=fileDescription, width=labelwidth, anchor=W, font=label_font, bg=label_bg_clr)
        self.lblName.grid(row=0, column=0)
        if not entry_box_width:
            self.entPath = Entry(self, textvariable=self.filePath, state=self.status,  validate="key", validatecommand=self.validation_methods.get(validation, None), font=entry_font, justify=justify)
        else:
            self.entPath = Entry(self, textvariable=self.filePath, state=self.status, width=entry_box_width, validate="key", font=entry_font, justify=justify, validatecommand=self.validation_methods.get(validation, None))
        self.entPath.grid(row=0, column=1)
        if value is not None:
            self.entry_set(val=value)

    @property
    def entry_get(self):
        return self.entPath.get()

    def entry_set(self, val):
        self.filePath.set(val)

    def set_state(self, setstatus):
        self.entPath.config(state=setstatus)

    def destroy(self):
        try:
            self.lblName.destroy()
            self.entPath.destroy()
        except:
            pass


class FolderSelect(Frame):
    def __init__(self,
                 parent: Union[Frame, LabelFrame, Canvas, Toplevel],
                 folderDescription: Optional[str] = "",
                 color: Optional[str] = None,
                 label_bg_clr: Optional[str] = None,
                 font: Tuple = Formats.FONT_REGULAR.value,
                 title: Optional[str] = "",
                 lblwidth: Optional[int] = 0,
                 bg_clr: Optional[str] = 'white',
                 entry_width: Optional[int] = 20,
                 initialdir: Optional[Union[str, os.PathLike]] = None,
                 tooltip_txt: Optional[str] = None,
                 **kw):

        self.title, self.initialdir = title, initialdir
        self.color = color if color is not None else "black"
        self.lblwidth = lblwidth if lblwidth is not None else 0
        self.parent = parent
        Frame.__init__(self, master=parent, **kw)
        browse_icon = ImageTk.PhotoImage(image=PIL.Image.open(MENU_ICONS["browse_small"]["icon_path"]))
        self.folderPath = StringVar()
        self.lblName = Label(self, text=folderDescription, fg=str(self.color), width=str(self.lblwidth), anchor=W, font=font, bg=label_bg_clr)
        self.lblName.grid(row=0, column=0, sticky=NW)
        self.entPath = Label(self, textvariable=self.folderPath, relief=SUNKEN, font=font, bg=bg_clr, width=entry_width)
        self.entPath.grid(row=0, column=1, sticky=NW)
        self.btnFind = SimbaButton(parent=self, txt=Defaults.BROWSE_FOLDER_BTN_TEXT.value, font=font, cmd=self.setFolderPath, img=browse_icon)
        #self.btnFind = Button(self, text=Defaults.BROWSE_FOLDER_BTN_TEXT.value, compound="left", image=browse_icon, relief=RAISED, font=font, command=self.setFolderPath)
        self.btnFind.image = browse_icon
        self.btnFind.grid(row=0, column=2, sticky=NW)
        self.folderPath.set("No folder selected")

        if tooltip_txt is not None and isinstance(tooltip_txt, str):
            CreateToolTip(widget=self.entPath, text=tooltip_txt)

    def setFolderPath(self):
        if self.initialdir is not None:
            if not os.path.isdir(self.initialdir):
                self.initialdir = None
            else:
                pass
        folder_selected = askdirectory(title=str(self.title), parent=self.parent, initialdir=self.initialdir)
        if folder_selected:
            self.folderPath.set(folder_selected)
            self.entPath.configure(width=len(folder_selected)+10)
        else:
            self.folderPath.set("No folder selected")
            self.entPath.configure(width=len("No folder selected")+10)


    @property
    def folder_path(self):
        return self.folderPath.get()


class ToolTip(object):

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 20
        y = y + cy + self.widget.winfo_rooty() + 20
        self.tipwindow = tw = Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = Label(tw, text=self.text, justify=LEFT, background="#ffffe0", relief=SOLID, borderwidth=1, font=Formats.FONT_REGULAR.value)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


def CreateToolTip(widget, text):
    toolTip = ToolTip(widget)
    def enter(event):
        toolTip.showtip(text)

    def leave(event):
        toolTip.hidetip()

    widget.bind("<Enter>", enter)
    widget.bind("<Leave>", leave)


def CreateLabelFrameWithIcon(parent: Union[Toplevel, LabelFrame, Canvas, Frame],
                             header: str,
                             icon_name: str,
                             padx: Optional[int] = None,
                             pady: Optional[int] = None,
                             relief: str = 'solid',
                             width: Optional[int] = None,
                             bg: Optional[str] = None,
                             font: tuple = Formats.FONT_HEADER.value,
                             icon_link: Optional[Union[str, None]] = LabelFrame):

    icon = PIL.Image.open(MENU_ICONS[icon_name]["icon_path"])
    icon = ImageTk.PhotoImage(icon)

    frm = Frame(parent, bg=bg)
    label_text = Label(frm, text=header, font=font, bg=bg)
    label_text.grid(row=0, column=0)

    label_image = Label(frm, image=icon, bg=bg)
    label_image.image = icon
    if icon_link:
        label_image.bind("<Button-1>", lambda e: callback(icon_link))
    label_image.grid(row=0, column=1)

    lbl_frm = LabelFrame(parent, labelwidget=frm, relief=relief, width=width, padx=padx or 0, pady=pady or 0)
    if bg is not None:
        lbl_frm.configure(bg=bg)
    return lbl_frm

def callback(url):
    webbrowser.open_new(url)


def create_scalebar(
    parent: Frame, name: str, min: int, max: int, cmd: object or None = None
):

    scale = Scale(
        parent,
        from_=min,
        to=max,
        orient=HORIZONTAL,
        length=200,
        label=name,
        command=cmd,
    )
    scale.set(0)
    return scale


class TwoOptionQuestionPopUp(object):
    """
    Helpe to create a two-option question tkinter pop up window (e.g., YES/NO).

    :parameter str question: Question to present to the user. E.g., ``DO YOU WANT TO PROCEED?``.
    :parameter str option_one: The first user option. E.g., ``YES``.
    :parameter str option_one: The second user option. E.g., ``NO``.
    :parameter str title: The window titel in the top banner of the pop-up. E.g., ``A QUESTION FOR YOU!``.
    :parameter Optional[str] link: If not None, then a link to documentation presenting background info about the user choices.
    """

    def __init__(self,
                 question: str,
                 option_one: str,
                 option_two: str,
                 title: str,
                 link: Optional[str] = None):

        self.main_frm = Toplevel()
        self.main_frm.geometry("600x200")
        self.main_frm.title(title)

        question_frm = Frame(self.main_frm)
        question_frm.pack(expand=True, fill="both")
        Label(question_frm, text=question, font=Formats.LABELFRAME_HEADER_FORMAT.value).pack()

        button_one = SimbaButton(parent=question_frm, txt=option_one, txt_clr="blue", bg_clr="lightgrey", img='check_blue', cmd=self.run, cmd_kwargs={'selected_option': lambda: option_one}, font=Formats.FONT_LARGE.value)
        button_two = SimbaButton(parent=question_frm, txt=option_two, txt_clr="red", bg_clr="lightgrey", img='close', cmd=self.run, cmd_kwargs={'selected_option': lambda: option_two}, font=Formats.FONT_LARGE.value)
        #button_one = Button(question_frm, text=option_one, fg="blue", command=lambda: self.run(option_one))
        #button_two = Button(question_frm, text=option_two, fg="red", command=lambda: self.run(option_two))
        if link:
            link_lbl = Label(question_frm, text="Click here for more information.", cursor="hand2", fg="blue")
            link_lbl.bind("<Button-1>", lambda e: callback(link))
            link_lbl.place(relx=0.5, rely=0.30, anchor=CENTER)
        button_one.place(relx=0.5, rely=0.50, anchor=CENTER)
        button_two.place(relx=0.5, rely=0.80, anchor=CENTER)
        self.main_frm.wait_window()

    def run(self, selected_option):
        self.selected_option = selected_option
        self.main_frm.destroy()

def SimbaButton(parent: Union[Frame, Canvas, LabelFrame, Toplevel],
                txt: str,
                txt_clr: Optional[str] = 'black',
                bg_clr: Optional[str] = '#f0f0f0',
                active_bg_clr: Optional[str] = None,
                hover_bg_clr: Optional[str] = Formats.BTN_HOVER_CLR.value,
                hover_font: Optional[Tuple] = Formats.FONT_REGULAR_BOLD.value,
                font: Optional[Tuple] = Formats.FONT_REGULAR.value,
                width: Optional[int] = None,
                height: Optional[int] = None,
                compound: Optional[str] = 'left',
                img: Optional[Union[ImageTk.PhotoImage, str]] = None,
                cmd: Optional[Callable] = None,
                cmd_kwargs: Optional[Dict[Any, Any]] = None,
                enabled: Optional[bool] = True,
                anchor: str = 'w',
                thread: Optional[bool] = False,
                tooltip_txt: Optional[str] = None) -> Button:
    def on_enter(e):
        e.widget.config(bg=hover_bg_clr, font=hover_font)

    def on_leave(e):
        e.widget.config(bg=bg_clr, font=font)

    if isinstance(img, str):
        img = ImageTk.PhotoImage(image=PIL.Image.open(MENU_ICONS[img]["icon_path"]))

    if cmd_kwargs is None:
        cmd_kwargs = {}

    def execute_command():
        if cmd:
            evaluated_kwargs = {k: (v() if callable(v) else v) for k, v in cmd_kwargs.items()}
            cmd(**evaluated_kwargs)

    if cmd is not None:
        if thread:
            command = lambda: threading.Thread(target=execute_command).start()
        else:
            command = execute_command
    else:
        command = None

    btn = Button(master=parent,
                 text=txt,
                 compound=compound,
                 image=img,
                 relief=RAISED,
                 fg=txt_clr,
                 activebackground=active_bg_clr,
                 font=font,
                 bg=bg_clr,
                 anchor=anchor,
                 command=command)

    if img is not None:
        btn.image = img
    if width is not None:
        btn.config(width=width)
    if height is not None:
        btn.config(height=height)
    if not enabled:
        btn.config(state=DISABLED)

    if hover_bg_clr is not None:
        btn.bind("<Enter>", on_enter)
        btn.bind("<Leave>", on_leave)

    if tooltip_txt is not None:
        CreateToolTip(widget=btn, text=tooltip_txt)

    return btn


def SimbaCheckbox(parent: Union[Frame, Toplevel, LabelFrame, Canvas],
                  txt: str,
                  txt_clr: Optional[str] = 'black',
                  txt_img: Optional[str] = None,
                  txt_img_location: Literal['left', 'right', 'top', 'bottom'] = RIGHT,
                  font: Optional[Tuple[str, str, int]] = Formats.FONT_REGULAR.value,
                  val: Optional[bool] = False,
                  state: Literal["disabled", 'normal'] = NORMAL,
                  cmd: Optional[Callable] = None,
                  tooltip_txt: Optional[str] = None):

    var = BooleanVar(value=False)
    if val: var.set(True)
    if isinstance(txt_img, str):
        txt_img = ImageTk.PhotoImage(image=PIL.Image.open(MENU_ICONS[txt_img]["icon_path"]))
    if cmd is None:
        cb = Checkbutton(master=parent, font=font, fg=txt_clr, image=txt_img, text=txt, compound=txt_img_location, variable=var)
    else:
        cb = Checkbutton(master=parent, font=font, fg=txt_clr, image=txt_img, text=txt, compound=txt_img_location, variable=var, command=cmd)
    if txt_img is not None:
        cb.image = txt_img
    if state == DISABLED:
        cb.configure(state=DISABLED)

    if isinstance(tooltip_txt, str):
        CreateToolTip(widget=cb, text=tooltip_txt)

    return cb, var

def SimBALabel(parent: Union[Frame, Canvas, LabelFrame, Toplevel],
               txt: str,
               txt_clr: str = 'black',
               bg_clr: Optional[str] = None,
               font: tuple = Formats.FONT_REGULAR.value,
               relief: str = FLAT,
               justify: Optional[str] = None,
               link: Optional[str] = None,
               width: Optional[int] = None,
               cursor: Optional[str] = None,
               anchor: Optional[str] = None):


    anchor = 'w' if anchor is None else anchor

    lbl = Label(parent,
                text=txt,
                font=font,
                fg=txt_clr,
                bg=bg_clr,
                justify=justify,
                relief=relief,
                cursor=cursor,
                anchor=anchor)

    if width is not None:
        lbl.configure(width=width)

    if link is not None:
        lbl.bind("<Button-1>", lambda e: callback(link))

    return lbl


def get_menu_icons():
    menu_icons = copy(MENU_ICONS)
    for k in menu_icons.keys():
        menu_icons[k]["img"] = ImageTk.PhotoImage(image=PIL.Image.open(os.path.join(os.path.dirname(__file__), menu_icons[k]["icon_path"])))
    return menu_icons


class SimBADropDown(Frame):
    def __init__(self,
                parent: Union[Frame, Canvas, LabelFrame, Toplevel, Tk],
                dropdown_options: Union[Iterable[Any], List[Any], Tuple[Any]],
                label: Optional[str] = None,
                label_width: Optional[int] = None,
                label_font: tuple = Formats.FONT_REGULAR.value,
                label_bg_clr: Optional[str] = None,
                dropdown_font_size: Optional[int] = None,
                justify: str = 'center',
                dropdown_width: Optional[int] = None,
                command: Callable = None,
                value: Optional[Any] = None,
                state: Optional[str] = None,
                tooltip_txt: Optional[str] = None):

        super().__init__(master=parent)
        self.dropdown_var = StringVar()
        self.dropdown_lbl = SimBALabel(parent=self, txt=label, txt_clr='black', bg_clr=label_bg_clr, font=label_font, width=label_width, anchor='w')
        #self.dropdown_lbl = Label(self, text=label, width=label_width, anchor="w", font=label_font, bg=label_bg_clr)
        self.dropdown_lbl.grid(row=0, column=0)
        self.dropdown_options = dropdown_options
        self.command = command
        if dropdown_font_size is None:
            drop_down_font = None
        else:
            drop_down_font = ("Poppins", dropdown_font_size)
        self.dropdown = Combobox(self, textvariable=self.dropdown_var, font=drop_down_font, values=self.dropdown_options, state="readonly", width=dropdown_width, justify=justify)
        self.dropdown.grid(row=0, column=1, sticky="nw")
        if value is not None: self.set_value(value=value)
        if command is not None:
            self.command = command
            self.dropdown.bind("<<ComboboxSelected>>", self.on_select)
        if state == 'disabled':
            self.disable()
        if isinstance(tooltip_txt, str):
            CreateToolTip(widget=self.dropdown_lbl, text=tooltip_txt)

    def set_value(self, value: Any):
        self.dropdown_var.set(value)

    def get_value(self):
        return self.dropdown_var.get()

    def enable(self):
        self.dropdown.configure(state="normal")

    def disable(self):
        self.dropdown.configure(state="disabled")

    def getChoices(self):
        return self.dropdown_var.get()

    def setChoices(self, choice):
        self.dropdown_var.set(choice)

    def on_select(self, event):
        selected_value = self.dropdown_var.get()
        self.command(selected_value)

    def set_width(self, width: int):
        self.dropdown.configure(width=width)

    def change_options(self, values, set_index, set_str):
        self.dropdown_var.set('')
        self.dropdown['values'] = values
        if isinstance(set_index, int) and (0 <= set_index <= len(values) - 1):
            self.dropdown_var.set(values[set_index])
        elif (set_str is not None) and (set_str in values):
            self.dropdown_var.set(set_str)
        else:
            self.dropdown_var.set(values[0])
        self.set_width(width=max(5, max(len(s) for s in values)))
        if self.dropdown['values'] == ('',) or len(values) == 0:
            self.disable()
        else:
            self.enable()


class DropDownMenu(Frame):
    def __init__(self,
                 parent=None,
                 dropdownLabel="",
                 choice_dict=None,
                 labelwidth="",
                 com=None,
                 val: Optional[Any] = None,
                 **kw):
        Frame.__init__(self, master=parent, **kw)
        self.dropdownvar = StringVar()
        self.lblName = Label(self, text=dropdownLabel, width=labelwidth, anchor=W, font=Formats.FONT_REGULAR.value)
        self.lblName.grid(row=0, column=0)
        self.choices = choice_dict
        self.popupMenu = OptionMenu(self, self.dropdownvar, *self.choices, command=com)
        self.popupMenu.grid(row=0, column=1)
        if val is not None:
            self.setChoices(val)
    def getChoices(self):
        return self.dropdownvar.get()
    def setChoices(self, choice):
        self.dropdownvar.set(choice)
    def enable(self):
        self.popupMenu.configure(state="normal")
    def disable(self):
        self.popupMenu.configure(state="disable")



class FileSelect(Frame):
    def __init__(self,
                 parent=None,
                 fileDescription="",
                 color=None,
                 title: Optional[str] = None,
                 lblwidth: Optional[int] = None,
                 file_types=None,
                 bg_clr: Optional[str] = 'white',
                 dropdown: Union[DropDownMenu, SimBADropDown] = None,
                 entry_width: Optional[int] = 20,
                 status: Optional[str] = None,
                 font: Tuple = Formats.FONT_REGULAR.value,
                 initialdir: Optional[Union[str, os.PathLike]] = None,
                 initial_path: Optional[Union[str, os.PathLike]] = None, **kw):

        self.title, self.dropdown, self.initialdir = title, dropdown, initialdir
        self.file_type = file_types
        self.color = color if color is not None else "black"
        self.lblwidth = lblwidth if lblwidth is not None else 0
        self.parent = parent
        Frame.__init__(self, master=parent, **kw)
        browse_icon = ImageTk.PhotoImage(image=PIL.Image.open(MENU_ICONS["browse"]["icon_path"]))
        self.filePath = StringVar()
        self.lblName = Label(self, text=fileDescription, fg=str(self.color), width=str(self.lblwidth), anchor=W, font=Formats.FONT_REGULAR.value)
        self.lblName.grid(row=0, column=0, sticky=W)
        self.entPath = Label(self, textvariable=self.filePath, relief=SUNKEN, font=Formats.FONT_REGULAR.value, bg=bg_clr, width=entry_width)
        self.entPath.grid(row=0, column=1)
        self.btnFind = SimbaButton(parent=self, txt=Defaults.BROWSE_FILE_BTN_TEXT.value, font=font, cmd=self.setFilePath, img=browse_icon)
        self.btnFind.grid(row=0, column=2)
        self.filePath.set(Defaults.NO_FILE_SELECTED_TEXT.value)
        if initial_path is not None:
            self.filePath.set(initial_path)
        if status is not None:
            self.set_state(setstatus=status)


    def setFilePath(self):
        if self.initialdir is not None:
            if not os.path.isdir(self.initialdir):
                self.initialdir = None
            else:
                pass

        if self.file_type:
            file_selected = askopenfilename(title=self.title, parent=self.parent, filetypes=self.file_type, initialdir=self.initialdir)
        else:
            file_selected = askopenfilename(title=self.title, parent=self.parent, initialdir=self.initialdir)
        if file_selected:
            if self.dropdown is not None:
                _, name, _ = get_fn_ext(filepath=file_selected)
                self.dropdown.setChoices(name)
                self.filePath.set(name)
            else:
                self.filePath.set(file_selected)
            self.entPath.configure(width=len(file_selected)+10)

        else:
            self.filePath.set(Defaults.NO_FILE_SELECTED_TEXT.value)
            self.entPath.configure(width=len(Defaults.NO_FILE_SELECTED_TEXT.value) + 10)

    @property
    def file_path(self):
        return self.filePath.get()

    def set_state(self, setstatus):
        self.entPath.config(state=setstatus)
        self.btnFind["state"] = setstatus





import os
from typing import Iterable, List, Optional, Union

import pandas as pd
from numba import typed

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import functools
import multiprocessing
from copy import deepcopy

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.image_mixin import ImageMixin
from simba.mixins.timeseries_features_mixin import TimeseriesFeatureMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_float, check_int,
    check_valid_array, check_valid_dataframe)
from simba.utils.enums import Defaults, Formats
from simba.utils.printing import SimbaTimer
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext, read_df,
                                    read_video_info, write_df)


class StraubTailAnalyzer(ConfigReader):
    """
    Class using background removed - egocentrically rotated - videos, to featurize tail behavior for downstream straub tail classification.

    :param Union[str, os.PathLike] config_path: Path to SimBA project_config.ini.
    :param Optional[Union[str, os.PathLike]] data_dir: Path to directory holding pose-estimation data. If None, the uses `project_folder/csv/outlier_corrected_movement_location` directory in SimBA project.
    :param Union[str, os.PathLike] video_dir: Path to directory holding videos. If None, the uses `project_folder/videos` directory in SimBA project.
    :param Union[str, os.PathLike] save_dir: Path to directory where to saved featurized pose-estimation data.
    :param Iterable[str] anchor_points: Iterable holding the names od the pose-estimated body-parts belonging to the tail.
    :param Iterable[str] body_parts: Iterable holding the names od the pose-estimated body-parts belonging to the animal hull.

    References
    ----------
    .. [1] Lazaro et al., Brainwide Genetic Capture for Conscious State Transitions, `biorxiv`, doi: https://doi.org/10.1101/2025.03.28.646066

    :example:
    >>> runner = StraubTailAnalyzer(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
    >>>                            data_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated',
    >>>                            video_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated',
    >>>                            save_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated\tail_features_additional',
    >>>                            anchor_points=('tail_base', 'tail_center', 'tail_tip'),
    >>>                            body_parts=('nose', 'left_ear', 'right_ear', 'right_side', 'left_side', 'tail_base'))
    >>> runner.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 anchor_points: Iterable[str],
                 body_parts: Iterable[str],
                 save_dir: Union[str, os.PathLike],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 video_dir: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True, create_logger=False)
        if data_dir is None:
            self.data_paths = find_files_of_filetypes_in_directory(directory=self.outlier_corrected_dir, extensions=['.csv'])
        else:
            self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'])
        if video_dir is not None:
            self.video_dir = video_dir
        self.paths = {}
        for data_path in self.data_paths:
            video = find_video_of_file(video_dir=self.video_dir, filename=get_fn_ext(filepath=data_path)[1])
            self.paths[data_path] = video
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        self.tail_cols, self.bp_cols = [], []
        for bp in anchor_points:
            self.tail_cols.append(f'{bp}_x'.lower()); self.tail_cols.append(f'{bp}_y'.lower())
        for bp in body_parts:
            self.bp_cols.append(f'{bp}_x'.lower()); self.bp_cols.append(f'{bp}_y'.lower())
        self.required_cols = self.tail_cols + self.bp_cols
        self.save_dir = save_dir

    def run(self):
        for file_cnt, (file_path, video_path) in enumerate(self.paths.items()):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            _, px_per_mm, fps = read_video_info(vid_info_df=self.video_info_df, video_name=video_name)
            print(f'Analyzing {video_name} ({file_cnt+1}/{len(self.data_paths)})...')
            save_path = os.path.join(self.save_dir, f'{video_name}.csv')
            print(video_path, save_path)
            if not os.path.isfile(save_path) and (video_path is not None) and os.path.isfile(video_path):
                df = read_df(file_path=file_path, file_type=self.file_type)
                out_df = deepcopy(df)
                df.columns = [str(x).lower() for x in df.columns]
                check_valid_dataframe(df=df, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.required_cols)

                tail_geometry_df = df[self.tail_cols].values.reshape(len(df), int(len(self.tail_cols) /2), 2).astype(np.int64)
                tail_geometries = GeometryMixin().bodyparts_to_polygon(data=tail_geometry_df, parallel_offset=35, pixels_per_mm=px_per_mm)

                hull_geometry_df = df[self.bp_cols].values.reshape(len(df), int(len(self.bp_cols) / 2), 2).astype(np.int64)
                hull_geometries = GeometryMixin().bodyparts_to_polygon(data=hull_geometry_df, parallel_offset=40, pixels_per_mm=px_per_mm)










                out_df['tail_area'] = GeometryMixin().multiframe_area(shapes=tail_geometries, pixels_per_mm=px_per_mm)

                tail_area_std = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=out_df['tail_area'].values.astype(np.float32), window_sizes=np.array([0.5, 1.0, 2.0]), sample_rate=fps, statistics=typed.List(['std']))
                out_df = pd.concat([out_df, pd.DataFrame(tail_area_std[0], columns=['tail_area_std_05', 'tail_area_std_1', 'tail_area_std_2'])], axis=1)

                tail_area_mean = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=out_df['tail_area'].values.astype(np.float32), window_sizes=np.array([0.5, 1.0, 2.0]), sample_rate=fps, statistics=typed.List(['mean']))
                out_df = pd.concat([out_df, pd.DataFrame(tail_area_mean[0], columns=['tail_area_mean_05', 'tail_area_mean_1', 'tail_area_mean_2'])], axis=1)

                tail_area_mad = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=out_df['tail_area'].values.astype(np.float32), window_sizes=np.array([0.5, 1.0, 2.0]), sample_rate=fps, statistics=typed.List(['mad']))
                out_df = pd.concat([out_df, pd.DataFrame(tail_area_mad[0], columns=['tail_area_mad_05', 'tail_area_mad_1', 'tail_area_mad_2'])], axis=1)

                out_df['tail_hausdorf_05'] = GeometryMixin().multiframe_hausdorff_distance(geometries=tail_geometries, lag=0.5, sample_rate=fps)
                out_df['tail_hausdorf_1'] = GeometryMixin().multiframe_hausdorff_distance(geometries=tail_geometries, lag=1, sample_rate=fps)
                out_df['tail_hausdorf_2'] = GeometryMixin().multiframe_hausdorff_distance(geometries=tail_geometries, lag=2, sample_rate=fps)

                out_df['hull_hausdorf_05'] = GeometryMixin().multiframe_hausdorff_distance(geometries=hull_geometries, lag=0.5, sample_rate=fps)
                out_df['hull_hausdorf_1'] = GeometryMixin().multiframe_hausdorff_distance(geometries=hull_geometries, lag=1, sample_rate=fps)
                out_df['hull_hausdorf_2'] = GeometryMixin().multiframe_hausdorff_distance(geometries=hull_geometries, lag=2, sample_rate=fps)

                if video_path is not None:
                    tail_imgs = ImageMixin().slice_shapes_in_imgs(imgs=video_path, shapes=tail_geometries, bg_color=(255, 255, 255), core_cnt=8, verbose=True)
                    tail_imgs = ImageMixin.pad_img_stack(image_dict=tail_imgs)
                    tail_imgs = np.stack(list(tail_imgs.values()))
                    out_df['tail_mse_05'] = ImageMixin.img_sliding_mse(imgs=tail_imgs, slide_length=0.5, sample_rate=float(fps))
                    out_df['tail_mse_1'] = ImageMixin.img_sliding_mse(imgs=tail_imgs, slide_length=1.0, sample_rate=float(fps))
                    out_df['tail_mse_2'] = ImageMixin.img_sliding_mse(imgs=tail_imgs, slide_length=2.0, sample_rate=float(fps))
                    video_timer.stop_timer()
                    write_df(df=out_df, file_type='csv', save_path=save_path)
                    print(video_timer.elapsed_time_str)

# runner = MitraTailAnalyzer(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                            data_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated',
#                            video_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated',
#                            save_dir=r'C:\troubleshooting\mitra\project_folder\videos\additional\bg_removed\rotated\tail_features_additional',
#                            anchor_points=('tail_base', 'tail_center', 'tail_tip'),
#                            body_parts=('nose', 'left_ear', 'right_ear', 'right_side', 'left_side', 'tail_base'))
# runner.run()


import json
import pandas as pd
import re
from datetime import datetime

DATA_PATH = r"C:\Users\sroni\Downloads\gitter_chat.json"
SAVE_PATH = r'C:\Users\sroni\OneDrive\Desktop\gitter_chat_stats.json'

with open(r"C:\Users\sroni\Downloads\gitter_chat.json", "r", encoding='utf-8') as f:
    data = json.load(f)

results = pd.DataFrame(columns=['USERNAME', 'BODY'])
for message in data['messages']:
    msg_sender = message['sender']
    msg_sender = msg_sender.replace('@', ' ')
    msg_sender = re.sub(r'\W+', '', msg_sender)
    if 'body' in message['content'].keys():
        body = message['content']['body']
        body = body.replace('\n', ' ')
        body = body.replace('\t', ' ')
    else:
        body = ' '
    body = re.sub(r'\W+', ' ', body)
    results.loc[len(results)] = [msg_sender, body]

post_cnt = len(results)
unique_posters = results['USERNAME'].nunique()
average_user_post_cnt = results['USERNAME'].value_counts().mean()

results = {'post_cnt': len(results), 'unique_users_with_posts': unique_posters, 'avg_posts_per_user': average_user_post_cnt, 'date': datetime.today().strftime('%Y-%m-%d')}


with open(SAVE_PATH, 'w') as f:
    json.dump(results, f, indent=4)
#results.to_csv(SAVE_PATH, index=None)





import os
import random
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.third_party_label_appenders.converters import \
    create_yolo_keypoint_yaml
from simba.utils.checks import (check_float, check_if_dir_exists, check_int,
                                check_str, check_valid_boolean,
                                check_valid_tuple)
from simba.utils.enums import Options
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_frm_of_video,
                                    read_sleap_h5)
from simba.utils.warnings import FrameRangeWarning
from simba.utils.yolo import keypoint_array_to_yolo_annotation_str


class SleapH52Yolo:
    """
    Convert SLEAP `.h5` pose estimation annotations to YOLO keypoint annotation format.

    Reads SLEAP `.h5` files and associated videos, samples frames based on a confidence
    threshold, extracts keypoints for one or more animals, and saves image-label pairs in a format
    compatible with YOLOv8 keypoint training.

    :param Union[str, os.PathLike] data_dir: Directory containing SLEAP `.h5` files.
    :param Union[str, os.PathLike] video_dir: Directory containing the videos associated with `.h5` files.
    :param Union[str, os.PathLike] save_dir: Directory to save YOLO-formatted images, labels, and metadata.
    :param Optional[int] frms_cnt: Number of frames to sample per video. If `None`, all valid frames are used.
    :param bool verbose: If True, print progress during processing.
    :param float threshold: Likelihood threshold below which poses are discarded.
    :param float train_size: Proportion of frames to assign to the training set (rest go to validation).
    :param Tuple[int, ...] flip_idx: Tuple indicating how to flip body-parts for augmentation. Length must match keypoint count.
    :param int animal_cnt: Number of animals tracked per frame.
    :param bool greyscale: If True, convert images to grayscale.
    :param bool clahe: If True, apply CLAHE (Contrast Limited Adaptive Histogram Equalization).
    :param float padding: Relative padding to apply around the bounding box of keypoints (range 0.0 to 1.0).
    :param Optional[str] single_id: Optional custom ID to assign all annotations the same class (used in single-animal datasets).


    :example:
    >>>DATA_DIR = r'D:\ares\data\termite_1\data'
    >>>VIDEO_DIR = r'D:\ares\data\termite_1\video'
    >>>SAVE_DIR = r"D:\imgs\sleap_h5"
    >>>runner = SleapH52Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, save_dir=SAVE_DIR, threshold=0.9, frms_cnt=50, single_id='termite')
    >>>runner.run()
    """

    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 video_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 frms_cnt: Optional[int] = None,
                 verbose: bool = True,
                 threshold: float = 0,
                 train_size: float = 0.7,
                 flip_idx: Tuple[int, ...] = None,
                 animal_cnt: int = 2,
                 greyscale: bool = False,
                 clahe: bool = False,
                 padding: float = 0.00,
                 single_id: Optional[str] = None):

        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.H5', '.h5'], as_dict=True, raise_error=True)
        self.video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, as_dict=True, raise_error=True)
        missing_video_paths = [x for x in self.video_paths.keys() if x not in self.data_paths.keys()]
        missing_data_paths = [x for x in self.data_paths.keys() if x not in self.video_paths.keys()]
        check_if_dir_exists(in_dir=save_dir)
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(save_dir, 'images', 'train'), os.path.join(save_dir, 'images', 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(save_dir, 'labels', 'train'), os.path.join(save_dir, 'labels', 'val')
        if flip_idx is not None: check_valid_tuple(x=flip_idx, source=f'{self.__class__.__name__} flip_idx', valid_dtypes=(int,), minimum_length=1)
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        check_int(name=f'{self.__class__.__name__} animal_cnt', value=animal_cnt, min_value=1)
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir])
        check_float(name=f'{self.__class__.__name__} instance_threshold', min_value=0.0, max_value=1.0,  raise_error=True, value=threshold)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_float(name=f'{self.__class__.__name__} padding', value=padding, max_value=1.0, min_value=0.0, raise_error=True)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        if frms_cnt is not None:
            check_int(name=f'{self.__class__.__name__} frms_cnt', value=frms_cnt, min_value=1, raise_error=True)
        if len(missing_video_paths) > 0:
            raise NoFilesFoundError(msg=f'Video(s) {missing_video_paths} could not be found in {video_dir} directory', source=self.__class__.__name__)
        if len(missing_data_paths) > 0:
            raise NoFilesFoundError(msg=f'CSV data for {missing_data_paths} could not be found in {data_dir} directory', source=self.__class__.__name__)
        self.map_dict = {v: f'animal_{k + 1}' for k, v in enumerate(range(animal_cnt))} if single_id is None else {0: single_id}
        self.threshold, self.frms_cnt, self.clahe, self.greyscale = threshold, frms_cnt, clahe, greyscale
        self.train_size, self.verbose, self.animal_cnt = train_size, verbose, animal_cnt
        self.padding, self.save_dir, self.flip_idx, self.single_id = padding, save_dir, flip_idx, single_id


    def run(self):
        dfs = []
        for file_cnt, (file_name, file_path) in enumerate(self.data_paths.items()):
            df = read_sleap_h5(file_path=file_path)
            p_cols = df.iloc[:, 2::3]
            df = df.iloc[p_cols[df.gt(self.threshold).all(axis=1)].index]
            df['frm_idx'] = df.index
            selected_frms = random.sample(list(df['frm_idx']), self.frms_cnt) if self.frms_cnt is not None else list(
                df['frm_idx'].unique())
            df = df[df['frm_idx'].isin(selected_frms)]
            df['video'] = self.video_paths[file_name]
            dfs.append(df)

        dfs, timer = pd.concat(dfs, axis=0), SimbaTimer(start=True)
        dfs['id'] = dfs['frm_idx'].astype(str) + dfs['video'].astype(str)
        train_idx = random.sample(list(dfs['id'].unique()), int(len(dfs['frm_idx'].unique()) * self.train_size))
        for frm_cnt, frm_id in enumerate(dfs['id'].unique()):
            frm_data = dfs[dfs['id'] == frm_id]
            video_path = list(frm_data['video'])[0]
            frm_idx = list(frm_data['frm_idx'])[0]
            video_meta = get_video_meta_data(video_path=video_path)
            if self.verbose:
                print(f'Processing frame: {frm_cnt + 1}/{len(dfs)} ...')
            if frm_idx > video_meta['frame_count']:
                FrameRangeWarning(msg=f'Frame {frm_idx} could not be read from video {video_path}. The video {video_meta["video_name"]} has {video_meta["frame_count"]} frames', source=self.__class__.__name__)
                continue
            img = read_frm_of_video(video_path=video_path, frame_index=frm_idx, greyscale=self.greyscale, clahe=self.clahe)
            img_h, img_w = img.shape[0], img.shape[1]
            if list(frm_data['id'])[0] in train_idx:
                img_save_path = os.path.join(self.img_train_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lbl_train_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            else:
                img_save_path = os.path.join(self.img_val_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lb_val_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            img_lbl = ''
            frm_data = frm_data.drop(['video', 'id', 'frm_idx'], axis=1).T.iloc[:, 0]
            animal_idxs = np.array_split(list(range(0, len(frm_data))), self.animal_cnt)
            for track_id, animal_idx in enumerate(animal_idxs):
                keypoints = frm_data[frm_data.index.isin(animal_idx)].values.reshape(-1, 3)
                keypoints[keypoints[:, 2] != 0.0, 2] = 2
                if frm_cnt == 0 and track_id == 0:
                    if self.flip_idx is not None and keypoints.shape[0] != len(self.flip_idx):
                        raise InvalidInputError(msg=f'The SLEAP data contains data for {keypoints.shape[0]} body-parts, but passed flip_idx suggests {len(self.flip_idx)} body-parts', source=self.__class__.__name__)
                    elif self.flip_idx is None:
                        self.flip_idx = tuple(list(range(0, keypoints.shape[0])))
                instance_str = f'{track_id} ' if self.single_id is None else f'0 '
                instance_str += keypoint_array_to_yolo_annotation_str(x=keypoints, img_w=img_w, img_h=img_h, padding=self.padding)
                img_lbl += instance_str
            with open(lbl_save_path, mode='wt', encoding='utf-8') as f:
                f.write(img_lbl)
            cv2.imwrite(img_save_path, img)
        create_yolo_keypoint_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=self.map_dict, save_path=self.map_path, kpt_shape=(len(self.flip_idx), 3), flip_idx=self.flip_idx)
        timer.stop_timer()
        stdout_success(msg=f'YOLO formated data saved in {self.save_dir} directory', source=self.__class__.__name__, elapsed_time=timer.elapsed_time_str)


# DATA_DIR = r'D:\ares\data\termite_1\data'
# VIDEO_DIR = r'D:\ares\data\termite_1\video'
# SAVE_DIR = r"D:\imgs\sleap_h5"
#
# runner = SleapH52Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, save_dir=SAVE_DIR, threshold=0.9, frms_cnt=50, single_id='termite')
# runner.run()
#df = read_sleap_h5(file_path=r"D:\ares\data\termite_1\termite.h5")




import os
from copy import deepcopy
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_if_df_field_is_boolean, check_if_dir_exists, check_instance,
    check_valid_boolean, check_valid_dataframe, check_valid_dict)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats
from simba.utils.errors import NoDataError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, read_video_info,
                                    str_2_bool)


class BooleanConditionalCalculator(ConfigReader):
    """
    Compute descriptive statistics (e.g., the time in seconds and number of frames) of multiple Boolean fields fullfilling user-defined conditions.

    For example, computedescriptive statistics for when Animal 1 is inside the shape Rectangle_1 while at the same time directing towards shape Polygon_1,
    while at the same time Animal 2 is outside shape Rectangle_1 and directing towards Polygon_1.

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format.
    :param Dict[str, Union[bool, str]] rules: Rules with field names as keys and bools (or string representations of bools) as values.
    :param Optional[Union[str, os.PathLike, None]] data_path: Optional data paths to be processsed. Can be a directory or file path. If None, all CSVs inside the `projecet_folder/csv/outlier_corrected_movement_location` are analysed.
    :param Optional[Union[str, os.PathLike]] agg_save_path: Optional location where to save the aggregate results as CSV file. If None, then results are saved in project logs folder under the ``Detailed_conditional_aggregate_statistics_{self.datetime}.csv`` filename.
    :param Optional[Union[str, os.PathLike]] detailed_save_path: Optional location where to save the detailed results as CSV file (bout level data). If None, then results are saved in project logs folder under the ``Detailed_conditional_aggregate_statistics_{self.datetime}.csv`` filename.

    .. note:
       `Example expected aggregate output table <https://github.com/sgoldenlab/simba/blob/master/misc/Conditional_aggregate_statistics_20231004130314.csv>`__.
       `Example expected detailed output table <https://github.com/sgoldenlab/simba/blob/master/misc/Detailed_conditional_aggregate_statistics_20241011123409.csv>`__.

    :example I:
    >>> rules = {'Rectangle_1 Simon in zone': 'TRUE', 'Polygon_1 JJ in zone': 'TRUE'} #  OR {'Rectangle_1 Simon in zone': True, 'Polygon_1 JJ in zone': True}
    >>> conditional_bool_rule_calculator = BooleanConditionalCalculator(rules=rules, config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
    >>> conditional_bool_rule_calculator.run()
    >>> conditional_bool_rule_calculator.save()


    :example II:
    >>> rules = {'Stimulus 2 Animal_1 in zone': True, 'Stimulus 6 Animal_1 in zone': 'falsE'}
    >>> runner = BooleanConditionalCalculator(rules=rules, config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini", data_path=r'C:\troubleshooting\RAT_NOR\project_folder\csv\features_extracted')
    >>> runner.run()
    >>> runner.save()


    :references:
       .. [1] Shonka, Sophia, and Michael J Hylin. “Younger Is Better But Only for Males: Social Behavioral Development Following Juvenile Traumatic Brain Injury to the Prefrontal Cortex,” n.d.

    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 rules: Dict[str, Union[bool, str]],
                 data_path: Optional[Union[str, os.PathLike, None]] = None,
                 agg_save_path: Optional[Union[str, os.PathLike]] = None,
                 detailed_save_path: Optional[Union[str, os.PathLike]] = None,
                 verbose: bool = True):

        ConfigReader.__init__(self, config_path=config_path)
        check_instance(source=self.__class__.__name__, instance=rules, accepted_types=(dict,))
        check_valid_dict(x=rules, valid_key_dtypes=(str,), valid_values_dtypes=(str, bool,), min_len_keys=2, source=f'{self.__class__.__name__} rules')
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        if data_path is not None:
            if not os.path.isfile(data_path) and not os.path.isdir(data_path):
                raise NoDataError(msg=f'The data_path {data_path} is not a valid file-path or directory', source=self.__class__.__name__)
            elif os.path.isdir(data_path):
                self.data_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.file_type}'], as_dict=False, raise_error=False, raise_warning=True)
            else:
                self.data_paths = [data_path]
        else:
            data_path = self.features_dir
            self.data_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.file_type}'], as_dict=False, raise_error=False, raise_warning=True)
        if len(self.data_paths) == 0:
            raise NoDataError(msg=f'The data_path {data_path} has no valid data files', source=self.__class__.__name__)
        if agg_save_path is not None:
            check_if_dir_exists(in_dir=os.path.dirname(agg_save_path))
        else:
            agg_save_path = os.path.join(self.logs_path, f"Conditional_aggregate_statistics_{self.datetime}.csv")
        if detailed_save_path is not None:
            check_if_dir_exists(in_dir=os.path.dirname(detailed_save_path))
        else:
            detailed_save_path = os.path.join(self.logs_path, f"Detailed_conditional_aggregate_statistics_{self.datetime}.csv")
        self.agg_save_path, self.detailed_save_path, self.rules = agg_save_path, detailed_save_path, rules
        self.output_df = pd.DataFrame(columns=["VIDEO"] + list(self.rules.keys()) + ["TIME (s)", "FRAMES (count)"])
        self.bout_df_cols = ["VIDEO"] + list(self.rules.keys()) + ["START FRAME", "END FRAME", "START TIME", "END TIME" ,"BOUT TIME"]
        self.bout_dfs, self.rule_cols, self.verbose = [], list(self.rules.keys()), verbose
        self.rules = {k: str_2_bool(v) for k, v in self.rules.items()}


    def _slice_df(self, df: pd.DataFrame, rules: dict) -> pd.DataFrame:
        sliced_df = deepcopy(df)
        for k, v in rules.items():
            sliced_df = sliced_df[sliced_df[k] == 1] if v else sliced_df[sliced_df[k] == 0]
        return sliced_df



    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            if self.verbose: print(f'Analyzing conditional boolean statistics in {self.video_name}...({file_cnt+1}/{len(self.data_paths)})')
            _, _, self.fps = read_video_info(vid_info_df=self.video_info_df, video_name=self.video_name)
            self.df = read_df(file_path=file_path, file_type=self.file_type)
            check_valid_dataframe(df=self.df, source=file_path, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.rule_cols)
            for rule_col in self.rule_cols: check_if_df_field_is_boolean(df=self.df, field=rule_col, df_name=file_path)
            self.sliced_df = self._slice_df(df=self.df, rules=self.rules)
            time_s = round(len(self.sliced_df) / self.fps, 4)
            if len(self.sliced_df) > 0:
                bout_df = pd.DataFrame(data=np.zeros((len(self.df))), columns=['behavior'])
                bout_df.iloc[self.sliced_df.index] = 1
                bout_df = detect_bouts(data_df=bout_df, target_lst=['behavior'], fps=self.fps)
                bout_df = bout_df.assign(**{k: v for k, v in self.rules.items()})
                bout_df['VIDEO'] = self.video_name
                bout_df = bout_df.rename(columns={'Start_time': 'START TIME', 'End Time': 'END TIME', 'Start_frame': 'START FRAME', 'End_frame': 'END FRAME', 'Bout_time': 'BOUT TIME'})
                self.bout_dfs.append(bout_df[self.bout_df_cols])
            self.output_df.loc[len(self.output_df)] = ([self.video_name] + list(self.rules.values()) + [time_s] + [len(self.sliced_df)])


    def save(self):
        self.output_df.to_csv(self.agg_save_path, index=False)
        self.timer.stop_timer()
        stdout_success(msg=f"Boolean conditional data saved at at {self.agg_save_path}!", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)
        if len(self.bout_dfs) > 0:
            self.bout_dfs = pd.concat(self.bout_dfs, axis=0).reset_index(drop=True)
            self.bout_dfs.to_csv(self.detailed_save_path, index=False)
            stdout_success(msg=f"Detailed boolean conditional data saved at at {self.detailed_save_path}!", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


#'Stimulus 2 Animal_1 in zone', 'Stimulus 2 Animal_1 facing'
# rules = {'Stimulus 2 Animal_1 in zone': True, 'Stimulus 6 Animal_1 in zone': 'falsE'}
# runner = BooleanConditionalCalculator(rules=rules, config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini", data_path=r'C:\troubleshooting\RAT_NOR\project_folder\csv\features_extracted')
# runner.run()
# runner.save()
#






import os
from typing import Union

from simba.labelling.extract_labelling_meta import AnnotationMetaDataExtractor
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.utils.errors import CountError

VIDEO_SPLIT = 'INCLUDE ANNOTATION COUNTS SPLIT BY VIDEO'
BOUT_SPLIT = 'INCLUDE ANNOTATED BOUTS INFORMATION'

class ClfAnnotationCountPopUp(PopUpMixin, ConfigReader):
    """
    :example:
    >>> ClfAnnotationCountPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.config_path = config_path
        PopUpMixin.__init__(self, title='COUNT NUMBER OF ANNOTATIONS IN SIMBA PROJECT', config_path=config_path, icon='abacus')
        if len(self.clf_names) == 0:
            raise CountError(msg=f'No classifier names associated with SimBA project {config_path}', source=self.__class__.__name__)
        if len(self.target_file_paths) == 0:
            raise CountError(msg=f'No data files found inside the {self.targets_folder} directory. ' f'Cannot analyze annotation count without annotated data', source=self.__class__.__name__)
        self.settings_dict = self.create_cb_frame(cb_titles=[VIDEO_SPLIT, BOUT_SPLIT], main_frm=self.main_frm, frm_title='SETTINGS')
        self.create_run_frm(run_function=self.run, title='RUN')
        self.main_frm.mainloop()

    def run(self):
        split_by_video = self.settings_dict[VIDEO_SPLIT].get()
        annotated_bouts = self.settings_dict[BOUT_SPLIT].get()
        annotation_extractor = AnnotationMetaDataExtractor(config_path=self.config_path, split_by_video=split_by_video, annotated_bouts=annotated_bouts)
        annotation_extractor.run()
        annotation_extractor.save()